require 'net/http'
require 'net/https'
require 'uri'
require 'json'
require 'utility'
require 'digest'
require 'yaml'

class Spconfig < ActiveRecord::Base
  # Each spconfig can zero, one, or many owning groups, but we don't want
  # duplicates.
  has_many :groups

  belongs_to :federation

  # Each spconfig can zero, one, or many certificates. We want those
  # certificates to go away when the spconfig goes away.
  has_many :certificates, dependent: :destroy

  accepts_nested_attributes_for :groups,
                                reject_if:     :all_blank,
                                allow_destroy: true

  strip_attributes

  # We want a place to put the groups_attributes passed in from the
  # controller.
  attr_accessor :xgroups_attributes

  # We want to give admins the (hopefully rarely used) option to supress
  # the validity checks certain properties of the SP config.
  # attr_accessor :ignore_cert_issues
  attr_accessor :ignore_valid_until_issues
  attr_accessor :ignore_entityid_format_issues

  # We keep track of versions with paper_trail
  has_paper_trail

  ## VALIDATIONS
  validate  :metadata_is_valid
  validate  :at_least_one_owning_group
  validates :contact_email, presence: {message: 'You must supply a valid '\
                                                'mailing list address',}
  validates :entityid, uniqueness: {message: 'This entityID is already being '\
                                             'used by another configuration',},
                       if: -> { entityid.present? }

  ## CALLBACKS
  before_validation :get_validation_results

  before_destroy do
    # Uncomment the following to delete all versions of an spconfig object
    # when deleting the spconfig itself. Of course, doing that means you would
    # not be able to recover a deleted spconfig object from the versions table.
    # That is why we leave it commented.

    # self.versions.destroy_all
  end

  # We want to mark for export the metadata files after any create or delete.
  after_create  :after_create_callback
  after_destroy :after_destroy_callback

  # We want to strip out any Signatures before saving.
  before_save   :before_save_callback

  # We want to mark for export the metadata file after any save where there is
  # a metadata change.
  after_save    :after_save_callback

  # A few useful instance variables
  @metadata_validation_results = nil
  @sp_info                     = nil
  @saml_validation_errors      = nil

  # Here is where we define the mapping from metadata type to the basename
  # of its output file.
  @@metadata_type_to_basename = {
    'combined'   => 'metadata',
    'legacy'     => 'metadata-legacy',
    'non-legacy' => 'metadata-blanket',
  }
  # Each metadata file has a 'Name' parameter in the header and this name
  # must be unique for each metdata file consumed by the IdP. Thus, we
  # define a unique header name for each metadata file we output.
  @@metadata_type_to_headername = {
    'combined'   => 'urn:mace:stanford.edu:shibboleth:stanford',
    'legacy'     => 'https://spdb.stanford.edu/spmetadata/metadata-legacy.xml',
    'non-legacy' => 'https://spdb.stanford.edu/spmetadata/metadata-blanket.xml',
  }

  def federated?()
    if (self.federation) then
      return true
    else
      return false
    end
  end

  def self.log_me(message)
    Utility.logger.info "[model] #{message}"
  end

  def log_me(message)
    Utility.logger.info "[model] #{message}"
  end

  # Only call this when metadata changes
  def metadata_is_valid
    log_me('entering metadata_is_valid')

    # Go through the validation errors adding them as appropriate.
    if (!@saml_validation_errors.nil?)
      if (@saml_validation_errors.key?('200')) then
        # If the metadata contains '<?xml' skip the other errors.
        errors.add(:base, %q(Metadata must NOT contain the string '<?xml'))
        return
      elsif (@saml_validation_errors.key?('300')) then
        # If the metadata is not valid XML skip the other errors.
        errors.add(:base, 'Metadata is not valid XML')
        return
      elsif (@saml_validation_errors.key?('350')) then
        # If the metadata is not valid SAML XML skip the other errors.
        error_value = @saml_validation_errors['350'][0]
        error_value = error_value.strip()

        # Remove everything before the first colon (this gets rid of the
        # extra 'Metadata is not valid SAML 2.0 XML'.
        error_value.gsub!(/^[^:]*:/, '')

        errors.add(:base, "350: #{error_value}")
        return
      elsif (@saml_validation_errors.key?('370')) then
        # Contains a non-parsable certificate.
        errors.add(:base, 'Metadata contains a non-parsable certificate')
        return
      else
        # For the other errors, add them now. Note that we skip certificate
        # errors (5xx) and deal with them below.
        # Sort the errors.
        error_numbers = @saml_validation_errors.keys.sort
        error_numbers.each do |error_number|
          if (error_number =~ /^5/) then
            # This is a certificate error, so skip (for now).
            next
          end
          if (error_number == '410') then
            # This is a validUntil error we which we handle ourselves
            # below, so skip (for now).
            next
          end
          error_texts = @saml_validation_errors[error_number]
          error_texts.each do |error_text|
            errors.add(:base, "Metadata [#{error_number}]#{error_text}")
          end
        end
      end
    end

    # If the entityID is not defined, stop since other validation tests depend on
    # there being an entityID.
    if (not self.entityid) then
      errors.add(:base, 'Missing entityId.')
      return
    end

    # Deal with federated entries first.
    if (self.federated?) then
      log_me('this SP is part of an external federation')
      # Check that this entityId is in the incommon metdata. If not, add
      # error.
      if (not Utility.is_an_incommon_entity_id(self.entityid)) then
        errors.add(:base, "This entityId is not in my #{self.federation.name} metadata.")
      end

      log_me('has an associated federation so skipping metadata validation')
      # Since there is an associated federation, we skip all metadata validation.
      self.metadata = "SP metadata provided by the #{self.federation.name} federation"
      return
    else
      log_me('does not have an associated federation')
    end

    # Is the metadata associated with a federation?
    if (metadata =~ /SP metadata provided by the .* federation/) then
      log_me('metadata is from an external federation')
      metadata_from_external_federation = true
    else
      metadata_from_external_federation = false
    end

    # Skip validation of the metadata if this is an old record AND the
    # metadata field has NOT changed and this metadata is not from an
    # external federation.
    if (    !self.metadata_was.nil? \
        and !self.metadata_changed? \
        and !metadata_from_external_federation) then
      log_me('skipping validation as this is an old record ' \
             'and the metadata has not changed and is not from an external federation')
      return
    end

    # If the user has indicated that this SP is part of the InCommon Federation,
    # set metatadata to some default string and skip the rest of the validation.
    if (metadata.nil? or metadata.empty?) then
      errors.add(:base, 'Metadata cannot be empty.')
      return
    end

    # If the whitespace-trimmed version of the metadata from BEFORE the
    # edit agrees with the whitespace-trimmed version of the metadata from
    # AFTER the edit, then there is no reason to validate (unless the old metadata
    # was the string we put in for InCommon records)
    log_me('check to see if metadata has changed')
    if (self.metadata_was and
        (Utility.normalize_whitespace(self.metadata_was) == Utility.normalize_whitespace(metadata)) and
        (!metadata_from_external_federation)
       ) then
      return
    end

    length = @metadata_validation_results.length()

    if (length != 2) then
      raise 'metadata validator failed'
    end

    # Set the ignore_certificate_expiration and ignore_no_certificates
    # booleans based on the encryption_exceptions value.
    if ((not Rails.configuration.check_cert_expirations) or
        self.encryption_exception) then
      ignore_certificate_expiration = true
    else
      ignore_certificate_expiration = false
    end

    if ((not Rails.configuration.check_for_encrypting_cert) or
        self.encryption_exception) then
      ignore_no_encrypting_certificates = true
    else
      ignore_no_encrypting_certificates = false
    end

    if (self.ignore_entityid_format_issues.to_i() == 1) then
      ignore_entityid_format_issues = true
    else
      ignore_entityid_format_issues = false
    end

    # Set the ignore_valid_until_expiration boolean based on the
    # global Rails configuration and checkbox.
    if ((not Rails.configuration.check_valid_until_expirations) or
        (self.ignore_valid_until_issues.to_i() == 1)) then
      ignore_valid_until_expiration = true
    else
      ignore_valid_until_expiration = false
    end


    ## ## INCOMMON METADATA
    # Don't allow submission of metadata using an InCommon entityID (may
    # have to change this later).
    if (Utility.is_an_incommon_entity_id(self.entityid)) then
      msg = <<~MESSAGE.strip
        The entityID supplied ('#{self.entityid}')
        is an InCommon entityID; this SP's metadata
        is already known to our IdPs.
      MESSAGE
      errors.add(:base, msg)
    end

    ## ## CERTIFICATES

    # We need the certificates for the next couple of sectins.
    certificates = Utility.get_certificate_objects_from_metadata(metadata)

    # Get the counting hashes.
    number_certs_of_type = get_cert_use_count(certificates)

    log_me("\# certificates (encryption):   #{number_certs_of_type['encryption']}")
    log_me("\# certificates (signing):      #{number_certs_of_type['signing']}")
    log_me("\# certificates (unspecified):  #{number_certs_of_type['unspecified']}")
    log_me("\# certificates (unrecognized): #{number_certs_of_type['unrecognized']}")

    # The number of certificates suitable for encryption is the sum of the
    # number of certificates of type 'encryption' and type 'unspecified'
    # (remember that a certificate of type 'unspecified' can be used for
    # any purpose).
    number_certs_suitable_for_encryption = number_certs_of_type['encryption'] + \
                                           number_certs_of_type['unspecified']

    ## ## ENCRYPTING CERTIFICATE
    #
    # Unless we are ignoring the whole issue of encrypting certificates,
    # this validation checks that at least one of the certificates can be
    # used for encryption. Note that this validation does not care whether
    # the certificate has expired or not.
    if (not ignore_no_encrypting_certificates) then

      if (number_certs_of_type['unrecognized'].positive?) then
        errors.add(:base, 'One of your certificates has an unrecognized use.')
      end

      exception_url = 'https://uit.stanford.edu/service/saml/exception'
      missing_cert_text = <<~MISSING_CERT_MESSAGE.strip
        Metadata must contain at least one certificate for encryption; if
        your application does not support encrypted responses you must file an
        exception; for more information see <a target="_blank" href="#{exception_url}">
        SAML Encryption Exception Process</a>.
      MISSING_CERT_MESSAGE
      if (certificates.empty? or number_certs_suitable_for_encryption.zero?) then
        errors.add(:base, missing_cert_text)
      end
    end

    # These next variables are used in a couple of the following sections
    # that check for various kinds of expirations. Because some SP's
    # generate a certificate for exactly one year, we can get submissions
    # that are a few hours short of one year. So, set our limit to be one
    # year minus 14 days.
    almost_one_year_of_seconds = 60 * 60 * 24 * (365 - 14)
    now_utc                    = Time.now.utc
    expiration_limit_utc_epoch = now_utc.to_i() + almost_one_year_of_seconds

    ## ## CERTIFICATE EXPIRATION
    #
    # Unless we are ignoring the whole issue of expired certificates, this
    # validation checks that at least one of the _encrypting_ certificates
    # is not expired or too close to expiring. If there _are_ no
    # certificates suitable for encrypting, we skip this validation.
    if ((not ignore_certificate_expiration) and \
        (not number_certs_suitable_for_encryption.zero?)) then

      # This is a hash mapping use type to the number of certificates
      # of that type that are not expired. Note that if no certificates
      # of that use type appear, there will be no key for that use type.
      suitable_for_encryption_count = 0

      # Check certificate expirations. For each KeyDescriptor, at least one
      # certificate must be valid for a year.
      certificates.each do |certificate|
        if (certificate.can_be_used_for_encryption?()) then
          # Is this certificate expiring far enough into the future?
          # NOT DONE HERE! MAKE SURE COMPARISON WORKS
          if (expiration_limit_utc_epoch < certificate.expires_at.utc.to_i) then
            # This certificate is OK, so we increment the counter that keeps track of
            # how many certificates of this usage type are still withing the expiration
            # window.
            suitable_for_encryption_count += 1
          else
            # This certificate is expired or is expiring too soon.
          end
        end
      end

      # Did we find one?
      if (suitable_for_encryption_count.zero?) then
        errors.add(:base,
                   'No certificate found with at least ' \
                   "one year left before expiration (use: #{usage_type})")
      end
    end

    ## ## VALIDUNTIL ATTRIBUTE
    # Check the validUntil attribute.
    # If the validUntil attribute exists and is less than our expiration limit raise a
    # validation error.
    if (not ignore_valid_until_expiration) then
      validUntil = self.valid_until
      if (not validUntil.blank?()) then
        valid_until_expiration_limit_utc = now_utc.to_i() +
                                           Rails.configuration.check_valid_until_expired_seconds
        if (validUntil.to_i() < valid_until_expiration_limit_utc) then
              expired_text = Rails.configuration.check_valid_until_expired_text
              errors.add(:base,
                         "Metadata's validUntil attribute (#{validUntil}) " \
                         "does not have at least #{expired_text} left before expiration")
        end
      end
    end

    ## EntityID blacklist
    # Make sure the submitted entityId does not any regular expression in
    # the entityID blacklist.
    Rails.configuration.entityid_blacklist.each do |entry_rx, entry_descriptor|
      if (self.entityid =~ /#{entry_rx}/) then
        errors.add(:base, %Q(Your entityID '#{self.entityid}' ) +
                          %Q(is reserved for '#{entry_descriptor}'))
      end
    end

    ## EntityID formatting.
    # For new records we require that the entityID be an absolute URI.
    # See also the "InCommon Metadata Registration Practice Statement"
    # page at https://www.incommon.org/federation/mrps/
    if ((not ignore_entityid_format_issues) and self.new_record?()) then
      # This is a NEW record
      # Is the entityID an absolute URI?
      validURI = true
      begin
        uri = URI.parse(self.entityid)
      rescue URI::InvalidURIError
        validURI = false
      end

      # Do our own check
      if (entityid !~ /^(http|https):\/\/.+/) then
        validURI = false
      end

      if (uri.host.nil?) then
        validURI = false
      end

      if (not validURI) then
        invalid_entityid_message = <<~INVALID_ENTITYID_MESSAGE.strip.html_safe
          Your entityID &quot;#{self.entityid}&quot; is not in the form of a URI. All
          SPDB records must have an entityID that is an absolute URI starting
          with &quot;http&quot; or &quot;https&quot;.
          See also <a target="_blank"
          href="https://www.incommon.org/federation/mrps/">InCommon
          Metadata Registration Practice Statement</a>.
        INVALID_ENTITYID_MESSAGE

        errors.add(:base, invalid_entityid_message)
      else
        # This is an EXISTING record so we skip the check.
      end
    end

  end

  # Function to add a parameter (used to pass information from controller
  # to model).
  def add_groups_attributes(groups_attributes)
    log_me('entering add_groups_attributes')
    if (not groups_attributes.nil?) then
      log_me("adding groups_attributes: #{groups_attributes.inspect}")
      self.xgroups_attributes = groups_attributes
    else
      log_me('groups_attributes is nil')
      self.xgroups_attributes = {}
    end
    return
  end

  # Check that there is at least one owning group AND that none
  def at_least_one_owning_group()
    log_me('entering at_least_one_owning_group')
    # remember that xgroups_attributes
    groups_attributes = self.xgroups_attributes
    log_me("xgroups_attributes is #{xgroups_attributes.inspect}")

    # Create an array of all workgroup names that are not to be destroyed
    not_destroyed_workgroups = []
    if (groups_attributes) then
      groups_attributes.each do |key, group_attr|
        # group_attr is a hash. At least one must have its key
        # '_destroy' not be 1.
        if (group_attr.key?('_destroy')) then
          destroy_value  = group_attr['_destroy']
          workgroup_name = group_attr['workgroup']
          log_me("workgroup_name: #{workgroup_name}, destroy_value: #{destroy_value}")
          if ((destroy_value != '1') and (destroy_value != 'true')) then
            not_destroyed_workgroups.push(workgroup_name)
          end
        end
      end
    end

    # If not_destroyed_workgroups is empty, there are no owning groups.
    # That is not allowed.
    if (not_destroyed_workgroups.empty?) then
      log_me('problem: not_destroyed_workgroups has zero length')
      errors.add(:base, 'Must have at least one owning group')
      return
    else
      log_me("not_destroyed_workgroups has length #{not_destroyed_workgroups.size}")
    end

    # Make sure at least one of the workgroups is not a personal
    # workgroup.
    at_least_one_workgroup_is_not_personal = false
    not_destroyed_workgroups.each do |workgroup_name|
      if (workgroup_name !~ /^~/) then
        # This is not a personal group
        at_least_one_workgroup_is_not_personal = true
        break
      end
    end

    if (not at_least_one_workgroup_is_not_personal) then
      errors.add(:base, 'Must have at least one non-personal owning group')
    end

    # If we get here, then everything is good and we can simply return.
    return
  end

  # Returns an a "counting" hash mapping usage type to number of
  # certificates of that type.
  #
  # The four types:
  #
  #  'signing'
  #  'encryption'
  #  'unspecified'
  #  'unrecognized'
  #
  # Example:
  #    { signing: 1, encryption: 1, unspecified: 0, unrecognized: 0 },
  #
  def get_cert_use_count(certificates)

    all_cert_use_count = {}
    all_cert_use_count['encryption']   = 0
    all_cert_use_count['signing']      = 0
    all_cert_use_count['unspecified']  = 0
    all_cert_use_count['unrecognized'] = 0

    certificates.each do |certificate|
      if (certificate.use_is_encryption?())
        usage_type = 'encryption'
      elsif (certificate.use_is_signing?())
        usage_type = 'signing'
      elsif (certificate.use_is_unspecified?())
        usage_type = 'unspecified'
      else
        usage_type = 'unrecognized'
      end

      all_cert_use_count[usage_type] += 1
    end

    return all_cert_use_count
  end

  def to_s()

    print "entityid:           #{self.entityid}\n"
    print "contact_email:      #{self.contact_email}\n"
    print "owned_by:           #{self.owned_by}\n"
    print "loginURL:           #{self.login_url}\n"
    print "created_at:         #{self.created_at}\n"
    print "updated_at:         #{self.updated_at}\n"
    print "created_by:         #{self.created_by}\n"
    print "updated_by:         #{self.updated_by}\n"
    print "sign_responses:     #{self.sign_responses}\n"
    print "sign_assertions:    #{self.sign_assertions}\n"
    print "encrypt_assertions: #{self.encrypt_assertions}\n"
  end


  # Write the metadata file (maybe). Return true if the file was written,
  # false otherwise. Skip any configurations with disabled field set to
  # 'true'.
  def self.write_metadata_file(metadata_type, force_update = false)

    # First of all, do we even _need_ to write the metadata file?
    if (not Spconfig.should_metadata_file_be_exported(metadata_type, force_update)) then
      log_me("no need to update metadata_type '#{metadata_type}' at this time")
      return false
    else
      log_me("metadata_type '#{metadata_type}' needs to be updated")
    end

    # If we get here then we need to write the file.

    # Step 1. Define the path where the output file will go.
    filepath = self.metadata_type_to_filepath(metadata_type)

    # f will be a temporary file for the complete metadata file.
    f = Tempfile.new(metadata_type)

    # Step 2. Create a raw file of all the metadata. This will vary depending
    # on the metadata_type.
    raw_metadata_file = Tempfile.new(metadata_type)
    raw_metadata_file.close()
    Spconfig.create_raw_metadata_file(metadata_type, raw_metadata_file)


    # Step 3. Construct the complete output file.

    # Step 3a. Make the header.
    # HEADER
    # We need to replace the string '%%VALID_UNTIL%%' with a valid until
    # time in form '2016-10-10T18:43:47Z'. How long this file is valid is
    # set as a number of seconds in the application.rb configuration file.
    header = Rails.configuration.sp_metadatafile_header.dup

    expiration_time    = Time.now.utc + Rails.configuration.sp_metadatafile_expire_seconds
    valid_until_string = expiration_time.strftime('%FT%TZ')

    header.sub!('%%VALID_UNTIL%%', valid_until_string)

    # We also need to replace the string '%%NAME%%' with the correct
    # headername (see the definition of @@metadata_type_to_headername
    # above for more details).
    headername = @@metadata_type_to_headername[metadata_type]
    header.sub!('%%NAME%%', headername)

    # We also need to replace '%%ID_REPLACE%%' with a unique
    # identifier. We use a timestamp not for any technical reason but just
    # to make it easier to determine when the metadata file was created.
    # Example:
    # ID='20161014075412MDrrr'. The 'MD' is the literal string 'MD' and
    # is there to indicate this is MetaData. The 'rrr' are three random
    # letters to make extra sure there is no ID name collision.
    current_time = Time.now.utc
    rrr          = (0...3).map { ('a'..'z').to_a[rand(26)] }.join
    id_string    = current_time.strftime('%Y%m%d%H%M%S') + 'MD' + rrr

    # Now do the replace.
    header.sub!('%%ID_REPLACE%%', id_string)

    f.write(header)
    f.write("\n")

    # Step 3b. Put in all the actual metadata.
    # MIDDLE (the raw metadata strings)
    File.open(raw_metadata_file, 'r') do |rf|
      rf.each_line do |line|
        f.write(line)
      end
    end

    # Step 3c. Put in the footer.
    # FOOTER
    f.write("\n")
    f.write(Rails.configuration.sp_metadatafile_footer)

    f.close()

    log_me("wrote complete unsigned metadata file to '#{f.path()}'")

    # Step 4. Remove the raw metadata file.
    # At this point, we can delete the temporary file raw_metadata
    raw_metadata_file.unlink()

    # Step 5. Sign the output file (maybe) and move the file into place.
    if (Rails.configuration.sp_metadatafile_sign) then
      stdout, stderr, status = Utility.sign_xml_file(f.path(), filepath)
      if (status.exitstatus != 0) then
        log_me("signing file error: [stdout: #{stdout}][stderr: #{stderr}][status: #{status}]")
      else
        log_me("signed metadata file and stored in '#{filepath}'")
      end
    else
      File.rename(f.path(), filepath)
      log_me("not signing metadata file; moved temporary unsigned metadata file to '#{filepath}'")
    end

    # Cleanup the temporary file
    if (File.exist?(f.path())) then
      f.unlink()
    end

    # We return true since the file was written.
    return true
  end

  def self.metadata_type_to_filepath(metadata_type)
    basepath = Rails.configuration.sp_metadatafile_basedir

    # Get the basename for this metadata type.
    basename = @@metadata_type_to_basename[metadata_type]

    basename_xml = "#{basename}.xml"
    filepath     = File.join(basepath, basename_xml)

    return filepath
  end

  # The metadata_type parameter must be one of these values:
  #    * 'combined'   - all the metadata
  #    * 'legacy'     - all the metadata whose is_legacy field is TRUE
  #    * 'non-legacy' - all the metadata whose is_legacy field is FALSE
  #
  def self.should_metadata_file_be_exported(
        metadata_type,
        force_update = false
      )
    metadata_filepath = Spconfig.metadata_type_to_filepath(metadata_type)

    return Utility.should_file_be_exported(
             filepath:           metadata_filepath,
             updatedtime_name:   'last-changed',
             force_update:       force_update,
             time_limit_seconds: 2 * 60 * 60,
             action_label:       'metadata-changes'
           )
  end


  def self.create_raw_metadata_file(metadata_type, filepath)
  # Read the metadata from all the records into a file in a normalized
  # order. Returns the path to the supplied filepath. Skip any configurations
  # with disabled field set to 'true'.
  #
  # Remember that we DON'T want to include metadata from an spconfig that is
  # part of a federation.
    if (metadata_type == 'combined') then
      spconfigs = Spconfig.where(federation_id: nil,
                                 disabled: false).order(:id).all()
    elsif (metadata_type == 'legacy') then
      spconfigs = Spconfig.where(federation_id: nil,
                                 is_legacy: true, disabled:false).order(:id).all()
    elsif (metadata_type == 'non-legacy') then
      spconfigs = Spconfig.where(federation_id: nil,
                                 is_legacy: false, disabled: false).order(:id).all()
    else
      raise "unrecognized metadata type #{metadata_type}"
    end

    File.open(filepath, 'w') do |f|
      spconfigs.each do |spc|
        f.write(spc.metadata)
      end
    end

    return filepath
  end

  # Save all the metadata_types to their respective files. We return a
  # hash that maps the metadata_type to a boolean. Mapping to true
  # means the file was actually written, false means the write was
  # skipped. Note that we skip all configurations with the disabled
  # field set to 'true'.
  def self.save_metadatas_to_file(force_update = false)
    metadata_type_to_was_written = {}
    @@metadata_type_to_basename.each do |metadata_type, basename|
      log_me("creating (if necessary) metadata output file for metadata_type '#{metadata_type}'")
      rv = Spconfig.write_metadata_file(metadata_type, force_update)
      metadata_type_to_was_written[metadata_type] = rv
    end
    return metadata_type_to_was_written
  end


  # Return true if user is in at least one of this object's groups
  def user_in_groups(user)
    groups.each do |group|
      log_me("checking to see if #{user} is in group #{group.workgroup}")
      if (Utility.user_member_of_privgroup_cached(user, group.workgroup))
        return true
      end
    end

    return false
  end

  # Return an array of the spconfigs that the given user has access
  # to. That is all the spconfigs that the user owns, or that contain a
  # group that the user is a member of.
  def self.this_users_spconfigs(user)
    # Step 1. Get all the spcongifs owned by this user.
    spconfigs_owned = Spconfig.where(owned_by: user)

    # Collect all the id's from @spconfigs_owned so that we can avoid
    # duplicating them in @spconfigs_group_owned.
    spconfigs_owned_ids = {}
    spconfigs_owned.each do |spconfig|
      spconfigs_owned_ids[spconfig.id] = true
    end

    # 2. Now we find all spconfigs that have an associated group
    # containing user as a member.
    spconfigs_group_owned = []

    spconfigs_all = Spconfig.all()
    spconfigs_all.each do |spconfig|
      if (spconfig.user_in_groups(user)) then
        if (not spconfigs_owned_ids.key?(spconfig.id)) then
          spconfigs_group_owned.push(spconfig)
        end
      end
    end

    spconfigs = spconfigs_owned + spconfigs_group_owned
    return spconfigs
  end

  def contact_is_stanford_address?()
    if (self.contact_email.nil?()) then
      return false
    end

    address = self.contact_email
    if (address =~ /^.*([.]|@)stanford[.]edu$/i) then
      return true
    else
      return false
    end
  end

  def contact_is_stanford_list_address?()
    if (self.contact_email.nil?()) then
      return false
    end

    address = self.contact_email
    if (address =~ /^.*@lists[.]stanford[.]edu$/i) then
      return true
    else
      return false
    end
  end

  # Return an array of the workgroup names. Will return the empty
  # array if there are none.
  def workgroup_names()
    if (groups.nil?() or groups.empty?())
      return []
    else
      return groups.map { |g| g.workgroup }
    end
  end

  def after_create_callback
    Spconfig.save_metadatas_to_file()
  end

  def after_destroy_callback
    Spconfig.save_metadatas_to_file()
    log_me('configuration deleted; touching the updatedtimes table')
    Updatedtime.touch()
  end

  # Call this method AFTER any save.
  def after_save_callback
    change_detected = false

    # Did the 'metadata' field change?
    if (not self.metadata_change.nil?) then
      change_detected = true
      log_me(%q(change in 'metadata' field detected))
    end

    # Did the 'is_legacy' field change?
    if (not self.is_legacy_change.nil?) then
      change_detected = true
      log_me(%q(change in 'is_legacy' field detected))
    end

    # Did the 'disabled' field change?
    if (not self.disabled_change.nil?) then
      change_detected = true
      log_me(%q(change in 'disabled' field detected))
    end

    if (change_detected) then
      log_me('XML-affecting change detected; touching the updatedtimes table')
      Updatedtime.touch()
    else
      log_me('No XML-affecting change detected; not touching the updatedtimes table')
    end

    # Remove any duplicate workgroups.
    workgroup_hash = {}
    self.groups.each do |group|
      workgroup = group.workgroup
      if (workgroup_hash.key?(workgroup)) then
        # This is already in the list, so delete
        group.delete()
        log_me("deleting duplicate workgroup '#{workgroup}'")
      else
        workgroup_hash[workgroup] = true
      end
    end

    # Add the certificates (if any)
    certificates = Utility.get_certificate_objects_from_metadata(metadata)

    self.certificates.destroy_all
    self.certificates << certificates
  end

  # Call this method BEFORE any save.
  def before_save_callback
    # Only need to call this method if there is no federation (remember:
    # beign part of a federation means no metadata).
    if (self.federation) then
      return
    end

    # 1. Strip any Signature elements
    begin
      stripped_string = Utility.strip_signatures(metadata)
    rescue
      log_me('error stripping signature')
      errors.add(:base, 'Error stripping signatures; maybe XML is bad?')
    else
      if (!stripped_string.nil?()) then
        # At least one Signature removed.
        self.metadata = stripped_string
      end
    end

    # 2. Remove the XML declaration <?xml ...> if there is one.
    self.metadata = try_to_strip_xml_declaration(self.metadata)
  end

  def entityid_md5sum()
    return Digest::MD5.hexdigest(self.entityid)
  end

  protected
  def get_validation_results
    # No validation needed if spconfig is a federation stub entry.
    if (self.federation) then
      log_me("entityid is #{self.entityid}")
      return
    end

    metadata_stripped = try_to_strip_xml_declaration(metadata)

    @metadata_validation_results = Utility.submit_metadata_for_validation(metadata_stripped)

    @sp_info                = @metadata_validation_results[0]
    @saml_validation_errors = @metadata_validation_results[1]

    log_me("metadata validation results: #{@metadata_validation_results.inspect}")
    # Pre-populate entityid so we we can find it.
    if (@sp_info.key?('entityID'))
      self.entityid = @sp_info['entityID']
    end

    # Also populate validUntil (if it exists)
    if (@sp_info.key?('validUntil') and !@sp_info['validUntil'].empty?()) then
      self.valid_until = DateTime.xmlschema(@sp_info['validUntil'])
    else
      self.valid_until = nil
    end
  end

  # Returns the XML without the <?xml header if the XML is valid. If the
  # XML is _not_ valid simply returns the original XML string.
  def try_to_strip_xml_declaration(metadata_string)
    stripped_metadata = nil
    begin
      stripped_metadata = Utility.strip_xml_declaration(metadata_string)
    rescue StandardError
      stripped_metadata = nil
    end

    if (stripped_metadata.nil?()) then
      return metadata_string
    else
      return stripped_metadata
    end
  end

end
